"""This is the clippy initialization file."""

# The general flow is as follows:
# Create the configurations (see comments in .config for details)
# Create the logger
from __future__ import annotations

import importlib
import logging

from .clippy_types import CLIPPY_CONFIG, AnyDict
from .config import _clippy_cfg

# Create the main configuraton object and expose it globally.
cfg = CLIPPY_CONFIG(_clippy_cfg)

# expose for serialization.
_dynamic_types: AnyDict = {}

# Set up logging
logfmt = logging.Formatter(cfg.get("logformat"))
logger = logging.getLogger(cfg.get("logname"))
handler = logging.StreamHandler()
handler.setFormatter(logfmt)
logger.addHandler(handler)
logger.setLevel(cfg.get("loglevel"))


def load_classes():
    """For each listed backend, import the module of the same name. The
    backend should expose two functions: a classes() function that returns
    a dictionary of classes keyed by name, and a get_cfg() function that
    returns a CLIPPY_CONFIG object with backend-specific configuration.
    This object is then made an attribute of the global configuration
    (i.e., `cfg.fs.get('fs_specific_config')`).
    """
    for backend in cfg.get("backends"):
        b = importlib.import_module(f".backends.{backend}", package=__name__)
        setattr(cfg, backend, b.get_cfg())
        for name, c in b.classes().items():
            # backend_config = importlib.import_module(f".backends.{name}.config.{name}_config")
            # for k, v in backend_config.items():
            #     cfg._set(k, v)

            globals()[name] = c


load_classes()



"""Filesystem backend for clippy."""

from __future__ import annotations

import json
import logging
import os
import pathlib
import stat
import sys
from subprocess import CalledProcessError
from typing import Any

from ... import constants as clippy_constants
from ...clippy_types import CLIPPY_CONFIG
from ...error import (
    ClippyConfigurationError,
    ClippyInvalidSelectorError,
    ClippyTypeError,
    ClippyValidationError,
)
from ...selectors import Selector
from ...utils import flat_dict_to_nested
from ..serialization import ClippySerializable
from ..version import _check_version
from .config import _fs_config_entries
from .execution import _help, _run, _validate

# create a fs-specific configuration.
cfg = CLIPPY_CONFIG(_fs_config_entries)

PATH = sys.path[0]


def _is_user_executable(path: pathlib.Path) -> bool:
    # Must be a regular file
    if not os.path.isfile(path):
        return False

    st = os.stat(path)
    mode = st.st_mode
    uid = os.getuid()
    gid = os.getgid()

    # Owner permissions
    if st.st_uid == uid and mode & stat.S_IXUSR:  # noqa: SIM114
        return True
    # Group permissions
    elif st.st_gid == gid and mode & stat.S_IXGRP:  # noqa: SIM114
        return True
    # Other permissions
    elif mode & stat.S_IXOTH:
        return True

    return False


def get_cfg() -> CLIPPY_CONFIG:
    """This is a mandatory function for all backends. It returns the backend-specific configuration."""
    return cfg


def classes() -> dict[str, Any]:
    """This is a mandatory function for all backends. It returns a dictionary of class name
    to the actual Class for all classes supported by the backend."""
    from ... import cfg as topcfg  # pylint: disable=import-outside-toplevel

    paths = cfg.get("fs_backend_paths")
    _classes = {}
    # iterate over all (filesystem) paths and for those that aren't explicitly excluded,
    # create the class based on the directory name and add methods based on the executables
    # within the directory.
    for path in paths:
        files = os.scandir(path)
        for f in files:
            if f.name in cfg.get("fs_exclude_paths"):
                continue
            p = pathlib.Path(path, f)
            if os.path.isdir(p):
                _cls = _create_class(f.name, path, topcfg)
                _classes[f.name] = _cls
    return _classes


def _create_class(name: str, path: str, topcfg: CLIPPY_CONFIG):
    """Given a name, a path, and a master configuration, create
    a class with the given name, and add methods based on the
    executable files within the class. Set convenience fields
    (_name, _path, _cfg) as well, and load class metadata from
    a meta.json file in each class directory. The meta.json
    file typically holds the class's docstring and any initial
    top-level selectors as a dictionary of selector: docstring."""
    metafile = pathlib.Path(path, name, clippy_constants.CLASS_META_FILE)
    meta = {}
    if metafile.exists():
        with open(metafile, encoding="utf-8") as json_file:
            meta = json.load(json_file)
    # pull the selectors out since we don't want them in the class definition right now
    selectors = meta.pop(clippy_constants.INITIAL_SELECTOR_KEY, {})
    meta["_name"] = name
    meta["_path"] = path
    meta["_cfg"] = topcfg
    class_logger = logging.getLogger(topcfg.get("logname") + "." + name)
    class_logger.setLevel(topcfg.get("loglevel"))
    meta["logger"] = class_logger

    cls = type(name, (ClippySerializable,), meta)
    classpath = pathlib.Path(path, name)
    for file in os.scandir(classpath):
        fullpath = pathlib.Path(classpath, file)
        if _is_user_executable(fullpath):
            try:
                _process_executable(str(fullpath), cls)
            except ClippyConfigurationError as e:
                class_logger.warning("error processing %s: %s; ignoring", fullpath, e)

    # add the selectors
    # this should be in the meta.json file.
    for selector, docstr in selectors.items():
        class_logger.debug("adding %s to class", selector)
        setattr(cls, selector, Selector(None, selector, docstr))
    return cls


def _process_executable(executable: str, cls):
    """
    Stores the executable as a method of cls.
    details: The executable is queried for its description. The returned
             json file is parsed and converted to a type representing
             the executable's methods and state. The type is recorded in the
             symtable (by default globals()).
    """

    # grab the help string, which gives us the docstring and  all arguments.
    cls.logger.debug("processing executable %s", executable)
    try:
        j = _help(executable, {}, cls.logger)

    except CalledProcessError as e:
        raise ClippyConfigurationError("Execution error " + e.stderr) from e

    # check to make sure we have the method name. This is so the executable can have
    # a different name than the actual method.
    if clippy_constants.METHODNAME_KEY not in j:
        raise ClippyConfigurationError("No method_name in " + executable)
    # check version
    if not _check_version(j):
        raise ClippyConfigurationError("Invalid version information in " + executable)

    docstring = j.get(clippy_constants.DOCSTRING_KEY, "")
    args: dict[str, dict] = j.get(clippy_constants.ARGS_KEY, {})  # this is now a dict.

    # Create a list of descriptions ordered by position (excluding position=-1)
    ordered_descs = []
    unordered_descs = []
    for arg, arginfo in sorted(args.items(), key=lambda x: x[1]["position"]):
        desc = f"{arg}: {arginfo['desc']}"
        if arginfo["position"] != -1:
            ordered_descs.append(desc)
        else:
            desc += f", default: {arginfo['default_val']}"
            unordered_descs.append(desc)

    if ordered_descs != []:
        docstring += "\nRequired Arguments:\n"
        for desc in ordered_descs:
            docstring += f"  {desc}\n"

    if unordered_descs != []:
        docstring += "\nOptional Arguments:\n"
        for desc in unordered_descs:
            docstring += f"  {desc}\n"

    # if we don't explicitly pass the method name, use the name of the exe.
    method = j.get(clippy_constants.METHODNAME_KEY, os.path.basename(executable))
    if hasattr(cls, method) and not method.startswith("__"):
        cls.logger.warning(f"Overwriting existing method {method} for class {cls} with executable {executable}")
    _define_method(cls, method, executable, docstring, args)
    return cls


def _define_method(cls, name: str, executable: str, docstr: str, arguments: dict[str, dict] | None):  # pylint: disable=too-complex
    """Defines a method on a given class."""
    if arguments is None:
        arguments = {}

    def m(self, *args, **kwargs):
        """
        Generic Method that calls an executable with specified arguments
        """

        # special cases for __init__
        # call the superclass to initialize the _state
        if name == "__init__":
            super(cls, self).__init__()

        argdict = {}
        # statej  = {}

        # make json from args and state

        # .. add state
        # argdict[STATE_KEY] = self._state
        argdict[clippy_constants.STATE_KEY] = getattr(self, clippy_constants.STATE_KEY)
        # ~ for key in statedesc:
        #     ~ statej[key] = getattr(self, key)

        # .. add positional arguments
        numpositionals = len(args)
        for argdesc in arguments:
            value = arguments[argdesc]
            if 0 <= value.get("position", -1) < numpositionals:
                argdict[argdesc] = args[value["position"]]

        # .. add keyword arguments
        argdict.update(kwargs)

        # validate
        valid, stderr = _validate(executable, argdict, self.logger)
        if not valid:
            raise ClippyValidationError(stderr)

        # call executable and create json output
        
        outj = _run(executable, argdict, self.logger)

        # if we have results that have keys that are in our
        # kwargs, let's update the kwarg references. Works
        # for lists and dicts only.
        for kw, kwval in kwargs.items():
            if kw in outj.get(clippy_constants.REFERENCE_KEY, {}):
                kwval.clear()
                if isinstance(kwval, dict):
                    kwval.update(outj[kw])
                elif isinstance(kwval, list):
                    kwval += outj[kw]
                else:
                    raise ClippyTypeError()

        # dump any output
        if clippy_constants.OUTPUT_KEY in outj:
            print(outj[clippy_constants.OUTPUT_KEY])
        # update state according to json output
        if clippy_constants.STATE_KEY in outj:
            setattr(self, clippy_constants.STATE_KEY, outj[clippy_constants.STATE_KEY])

        # update selectors if necessary.
        if clippy_constants.SELECTOR_KEY in outj:
            d = flat_dict_to_nested(outj[clippy_constants.SELECTOR_KEY])
            for topsel, subsels in d.items():
                if not hasattr(self, topsel):
                    raise ClippyInvalidSelectorError(f"selector {topsel} not found in class; aborting")
                getattr(self, topsel)._import_from_dict(subsels)

        # return result
        if outj.get(clippy_constants.SELF_KEY, False):
            return self
        return outj.get(clippy_constants.RETURN_KEY)

        # end of nested def m

    # Add a new member function with name and implementation m to the class cls
    # setattr(name, '__doc__', docstr)
    m.__doc__ = docstr
    setattr(cls, name, m)



"""
Functions to check versions. Requires semver.
"""

from __future__ import annotations

from semver import Version

from .. import cfg
from ..clippy_types import AnyDict


def _check_version(output_dict: AnyDict | None) -> bool:
    """
    Given an output dictionary, match the version number against config.
    Return true if version is compatible with version range set in config,
    false otherwise. If no version is found, return false.
    """
    if output_dict is None:
        return False
    if "version" not in output_dict:
        return False
    backend_ver = Version.parse(output_dict["version"])
    config_vers = cfg.get("required_versions").split(",")
    return all(backend_ver.match(v.strip()) for v in config_vers)



"""
Clippy serialization functions and classes.
"""

from __future__ import annotations

from typing import Any

import jsonlogic as jl

from .. import _dynamic_types
from ..clippy_types import AnyDict
from ..error import ClippySerializationError

# TODO: SAB 20240204 complete typing here.


class ClippySerializable:
    """
    Declares the interface for serializing clippy objects. Subclass should inherit from this class to
    be serializable as part of the clippy frontend/backend communication infrustructure.

    The Clippy-type specification is as follows
    ```json
    {
        "__clippy_type__": {
            "__module__": <frontend-module>,
            "__class__": <clippy-type-name>,
            "state": {
                <state-attribute>: <state-attribute-value>
            }
        }
    }
    ```

    """

    def __init__(self):
        self._state = {}

    def to_serial(self) -> AnyDict:
        """
        Subclasses should override this method to provide a custom serialization of the object instance.
        This method should return a python-primitive (dict, list, str, etc) version of the instance state
        """
        return {
            "__clippy_type__": {
                "__module__": self.__class__.__module__,
                "__class__": self.__class__.__name__,
                "state": self._state,
            }
        }

    @classmethod
    def from_serial(cls, o: AnyDict):
        """
        Subclasses should override this method to provide a custom deserialization from the python-primative data.
        This method should return a fully constructed object instance.
        """

        if "__clippy_type__" not in o:
            raise ClippySerializationError("No clippy type detected")

        clippy_type = o["__clippy_type__"]
        assert isinstance(clippy_type, dict)
        # right now we aren't using the module but  probably should
        # type_module = clippy_type.get("__module__", None)
        type_name = clippy_type.get("__class__", None)
        state_dict = clippy_type.get("state", None)

        if type_name is None:
            raise ClippySerializationError("__clippy_type__.__class__ is unspecified")

        if type_name not in _dynamic_types:
            raise ClippySerializationError(f'"{type_name}" is not a known type, please clippy import it.')

        # get the type to deserialize into from the _dynamic_types dict
        # this does not account for the module the type may exist in
        t = _dynamic_types[type_name]

        if not issubclass(t, ClippySerializable):
            raise ClippySerializationError(f'"{type_name}" is not serializable.')

        # create an instance of the clippy type but avoid initializing it
        # because it may have required args we don't care about.
        # However we do want to call the ClippySerializable initializer
        # to give the instance the `state` attribute (and whatever else).
        instance = t.__new__(t)
        super(t, instance).__init__()

        # update the new instances state
        if state_dict is not None:
            instance._state.update(state_dict)
        return instance


def _form_method_arguments(method_args: AnyDict | None) -> tuple[list[Any], AnyDict]:
    if method_args is None:
        return [], {}

    sorted_method_args = sorted(
        [
            (
                method_args[arg_name]["position"],
                arg_name,
                method_args[arg_name]["arg_value"],
            )
            for arg_name in method_args
        ]
    )
    keyword_args = {arg[1]: arg[2] for arg in sorted_method_args if arg[0] == -1}
    positionals = [arg[2] for arg in sorted_method_args if arg[0] > -1]
    return (positionals, keyword_args)


def encode_clippy_json(o: Any) -> Any:
    """
    json encoder that is clippy-object aware.
    """
    if isinstance(o, jl.Operand):  # expression or variable
        return {"expression_type": "jsonlogic", "rule": o.prepare()}

    return o


def decode_clippy_json(o: Any) -> Any:
    """
    json decoder that is clippy-object aware.
    """
    if "__clippy_type__" in o:
        return ClippySerializable.from_serial(o)

    return o



# Copyright 2020 Lawrence Livermore National Security, LLC and other CLIPPy Project Developers.
# See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""This file contains custom Clippy errors."""


class ClippyError(Exception):
    """
    This is a top-level custom exception for Clippy.
    """


class ClippyConfigurationError(ClippyError):
    """
    This error represents a configuration error on user input.
    """


class ClippyBackendError(ClippyError):
    """
    This error should be thrown when the backend returns an abend.
    """


class ClippyValidationError(ClippyError):
    """
    This error represents a validation error in the inputs to a clippy job.
    """


class ClippySerializationError(ClippyError):
    """
    This error should be thrown when clippy object serialization fails
    """


class ClippyClassInconsistencyError(ClippyError):
    """
    This error represents a class inconsistency error (name or docstring mismatch).
    """


class ClippyTypeError(ClippyError):
    """
    This error represents an error with the type of data being passed to the back end.
    """


class ClippyInvalidSelectorError(ClippyError):
    """
    This error represents an error with a selector that is not defined for a given clippy class.
    """



"""Custom selectors for clippy."""

from __future__ import annotations

import jsonlogic as jl

from . import constants
from .clippy_types import AnyDict


class Selector(jl.Variable):
    """A Selector represents a single variable."""

    def __init__(self, parent: Selector | None, name: str, docstr: str):
        super().__init__(name, docstr)  # op and o2 are None to represent this as a variable.
        self.parent = parent
        self.name = name
        self.fullname: str = self.name if self.parent is None else f"{self.parent.fullname}.{self.name}"
        self.subselectors: set[Selector] = set()

    def __hash__(self):
        return hash(self.fullname)

    def prepare(self):
        return {"var": self.fullname}

    def hierarchy(self, acc: list[tuple[str, str]] | None = None):
        if acc is None:
            acc = []
        acc.append((self.fullname, self.__doc__ or ""))
        for subsel in self.subselectors:
            subsel.hierarchy(acc)
        return acc

    def describe(self):
        hier = self.hierarchy()
        maxlen = max(len(sub_desc[0]) for sub_desc in hier)
        return "\n".join(f"{sub_desc[0]:<{maxlen + 2}} {sub_desc[1]}" for sub_desc in hier)

    def __str__(self):
        return repr(self.prepare())

    def to_serial(self):
        return {"var": self.fullname}

    def _add_subselector(self, name: str, docstr: str):
        """add a subselector to this selector"""
        subsel = Selector(self, name, docstr)
        setattr(self, name, subsel)
        self.subselectors.add(subsel)

    def _del_subselector(self, name: str):
        delattr(self, name)
        self.subselectors.remove(getattr(self, name))

    def _clear_subselectors(self):
        """removes all subselectors"""
        for subsel in self.subselectors:
            delattr(self, subsel.name)
        self.subselectors = set()

    def _import_from_dict(self, d: AnyDict, merge: bool = False):
        """Imports subselectors from a dictionary.
        If `merge = True`, do not clear subselectors first.
        """

        # clear all children
        if not merge:
            self._clear_subselectors()

        for name, subdict in d.get(constants.SELECTOR_KEY, {}).items():
            docstr = subdict.get("__doc__", "")
            self._add_subselector(name, docstr)
            getattr(self, name)._import_from_dict(subdict)



"""Clippy constants."""

# These should probably not be changed, which is why they're not in config.

# The name of the selector key in the meta.json files.
# these selectors cannot be dropped.
INITIAL_SELECTOR_KEY = "initial_selectors"

######
# Keys found in output from backend.
######
# The name of the selector key in the json returns.
# If this key exists in the results
# then the selector hierarchy is replaced with its (processed) contents.
SELECTOR_KEY = "_selectors"
# Any internal state kept in the Python frontend.
STATE_KEY = "_state"
# This will reference a string to be printed to stdout by the frontend.
OUTPUT_KEY = "stdout"
# This key is a boolean - if true, return self (to enable method chaining).
SELF_KEY = "returns_self"
# key to json entry that holds reference overrides from backend functions.
REFERENCE_KEY = "references"
# key to json entry that holds return data from backend functions.
RETURN_KEY = "returns"


# these keys are returned from the _help output when creating methods
# for classes.
METHODNAME_KEY = "method_name"
DOCSTRING_KEY = "desc"
ARGS_KEY = "args"


REAL = "real"
STRING = "string"
UINT = "uint"
INT = "int"
CLASS_META_FILE = "meta.json"



# Constants unique to the fs backend.
# the flag to pass for a dry run to make sure syntax is proper

DRY_RUN_FLAG = "--clippy-validate"
# the flag to pass to get detailed help for constructing the class
HELP_FLAG = "--clippy-help"



"""
Functions to execute backend programs.
"""

from __future__ import annotations

import contextlib
import json
import logging
import os
import select
import subprocess

from ... import cfg
from ...clippy_types import AnyDict
from ..serialization import decode_clippy_json, encode_clippy_json
from .constants import DRY_RUN_FLAG, HELP_FLAG


class NonZeroReturnCodeError(Exception):
    def __init__(self, execcmd, return_code, extramsg, message="returned non-zero exit code"):
        self.return_code = return_code
        self.execcmd = execcmd
        super().__init__(f"{execcmd} {message}: {return_code}\n{extramsg}")

def _stream_exec(
    cmd: list[str],
    submission_dict: AnyDict,
    logger: logging.Logger,
    validate: bool,
) -> tuple[AnyDict | None, str | None, int]:
    """
    Internal function.

    Executes the command specified with `execcmd` and
    passes `submission_dict` as JSON via STDIN.

    Logs debug messages with progress.
    Parses the object and returns a dictionary output.
    Returns the process result object, stderr, and the process return code.

    This function is used by _run and _validate. All options (pre_cmd and flags) should
    already be set.
    """

    logger.debug(f"Submission = {submission_dict}")
    # PP support passing objects
    # ~ cmd_stdin = json.dumps(submission_dict)
    cmd_stdin = json.dumps(submission_dict, default=encode_clippy_json)

    logger.debug("Calling %s with input %s", cmd, cmd_stdin)

    d = {}
    stderr_lines = []

    with subprocess.Popen(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf8"
    ) as proc:
        assert proc.stdin is not None
        assert proc.stdout is not None
        assert proc.stderr is not None

        proc.stdin.write(cmd_stdin + "\n")
        proc.stdin.flush()
        proc.stdin.close()

        progress = None
        # Use select with file descriptors and non-blocking reads
        stdout_fd = proc.stdout.fileno()
        stderr_fd = proc.stderr.fileno()
        fds = [stdout_fd, stderr_fd]

        # Set non-blocking mode
        os.set_blocking(stdout_fd, False)
        os.set_blocking(stderr_fd, False)

        # Line buffers for partial reads
        stdout_buffer = ""
        stderr_buffer = ""

        while fds or proc.poll() is None:
            # Only select if we have fds to monitor
            if fds:
                readable, _, _ = select.select(fds, [], [], 0.1)
            else:
                readable = []

            for fd in readable:
                try:
                    chunk = os.read(fd, 4096)
                    if not chunk:
                        # Stream closed (EOF)
                        fds.remove(fd)
                        continue

                    text = chunk.decode("utf-8")

                    if fd == stdout_fd:
                        stdout_buffer += text
                        while "\n" in stdout_buffer:
                            line, stdout_buffer = stdout_buffer.split("\n", 1)
                            d = json.loads(line, object_hook=decode_clippy_json)
                    elif fd == stderr_fd:
                        stderr_buffer += text
                        while "\n" in stderr_buffer:
                            line, stderr_buffer = stderr_buffer.split("\n", 1)
                            stderr_lines.append(line + "\n")
                            print(line, flush=True)
                except BlockingIOError:
                    # No data available right now
                    continue
                except OSError:
                    # Stream closed
                    if fd in fds:
                        fds.remove(fd)
                    continue

        # Process any remaining buffered data
        if stdout_buffer.strip():
            with contextlib.suppress(json.JSONDecodeError):
                d = json.loads(stdout_buffer, object_hook=decode_clippy_json)

        if stderr_buffer.strip():
            stderr_lines.append(stderr_buffer)
            print(stderr_buffer.rstrip(), flush=True)

    stderr = "".join(stderr_lines) if stderr_lines else None
    if progress is not None:
        progress.close()
    # if proc.returncode:
    #     raise (ClippyValidationError(stderr) if validate else ClippyBackendError(stderr))

    if not d:
        return None, stderr, proc.returncode
    if stderr:
        logger.debug("Received stderr: %s", stderr)
    if proc.returncode != 0:
        logger.debug("Process returned %d", proc.returncode)
    logger.debug("run(): final stdout = %s", d)

    return (d, stderr, proc.returncode)


def _validate(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> tuple[bool, str]:
    """
    Converts the dictionary dct into a json file and calls executable cmd with the DRY_RUN_FLAG.
    Returns True/False (validation successful) and any stderr.
    """

    if isinstance(cmd, str):
        cmd = [cmd]

    execcmd = cfg.get("validate_cmd_prefix").split() + cmd + [DRY_RUN_FLAG]
    logger.debug("Validating %s", cmd)

    _, stderr, retcode = _stream_exec(execcmd, dct, logger, validate=True)
    return retcode == 0, stderr or ""


def _run(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> AnyDict:
    """
    converts the dictionary dct into a json file and calls executable cmd. Prepends
    cmd_prefix configuration, if any.
    """

    if isinstance(cmd, str):
        cmd = [cmd]
    execcmd = cfg.get("cmd_prefix").split() + cmd
    logger.debug("Running %s", execcmd)
    # should we do something with stderr?

    output, stderr, retcode = _stream_exec(execcmd, dct, logger, validate=False)
    if retcode != 0:
        raise NonZeroReturnCodeError(execcmd, retcode, stderr)
    return output or {}


def _help(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> AnyDict:
    """
    Retrieves the help output from the clippy command. Prepends validate_cmd_prefix
    if set and appends HELP_FLAG.
    Unlike `_validate()`, does not append DRY_RUN_FLAG, and returns the output.
    """
    if isinstance(cmd, str):
        cmd = [cmd]
    execcmd = cfg.get("validate_cmd_prefix").split() + cmd + [HELP_FLAG]
    logger.debug("Running %s", execcmd)
    # should we do something with stderr?

    output, _, _ = _stream_exec(execcmd, dct, logger, validate=True)
    return output or {}



# pylint: disable=consider-using-namedtuple-or-dataclass
"""This holds a dictionary containing global configuration variables for clippy."""

# The format is config_key: (environment variable or None, default value)

import logging

from .clippy_types import CONFIG_ENTRY

_clippy_cfg: dict[str, CONFIG_ENTRY] = {
    # backends to use for Clippy.
    "backends": ("CLIPPY_BACKENDS", ["fs"]),
    # semver version restrictions for the backend
    "required_versions": ("CLIPPY_REQ_VERSIONS", ">=0.2.0"),
    # command prefix used to specify clippy task management with the HPC cluster
    # for instance, if using slurm this could be set to 'srun -n1 -ppdebug'
    "cmd_prefix": ("CLIPPY_CMD_PREFIX", ""),
    # command prefix used to specify clippy task management with the HPC cluster
    # for dry runs in certain environments. For instance, if using slurm this
    # could be set to 'srun -n1 -ppdebug'
    "validate_cmd_prefix": ("CLIPPY_VALIDATE_CMD_PREFIX", ""),
    # contol the log level of clippy
    "loglevel": ("CLIPPY_LOGLEVEL", logging.WARNING),
    "logformat": (
        "CLIPPY_LOGFORMAT",
        "%(asctime)s [%(filename)s:%(lineno)d (%(funcName)s) %(levelname)s: %(message)s",
    ),
    "logname": ("CLIPPY_LOGNAME", __name__),
}



"""
Utility functions
"""

from .clippy_types import AnyDict
from .constants import SELECTOR_KEY
from .error import ClippyInvalidSelectorError


def flat_dict_to_nested(input_dict: AnyDict) -> AnyDict:
    """input dictionary has dot-delineated keys which are then parsed as subkeys.
    That is: {'a.b.c': 5} becomes {'a': {'_selectors': {'b': {'_selectors': {'c': {'__doc__' = 5}}}}}}
    assuming `SELECTOR_KEY == '_selectors'`
    """

    output_dict: AnyDict = {}
    for k, v in input_dict.items():
        # k is dotted
        if "." not in k:
            raise ClippyInvalidSelectorError("cannot set top-level selectors")

        *path, last = k.split(".")
        if last.startswith("_"):
            raise ClippyInvalidSelectorError("selectors must not start with an underscore.")
        curr_nest = output_dict
        for p in path:
            if p.startswith("_"):
                raise ClippyInvalidSelectorError("selectors must not start with an underscore.")
            curr_nest.setdefault(p, {})
            curr_nest[p].setdefault(SELECTOR_KEY, {})
            curr_nest = curr_nest[p][SELECTOR_KEY]

        curr_nest.setdefault(last, {"__doc__": v})
    return output_dict



# pylint: disable=consider-using-namedtuple-or-dataclass
import os

_fs_config_entries = {
    # backend path for executables, in addition to the CLIPPY_BACKEND_PATH environment variable. Add to it here.
    "fs_backend_paths": (
        None,
        [x.strip() for x in os.environ.get("CLIPPY_BACKEND_PATH", "").split(":")],
    ),
    # exclude these directories from being made as classes.
    "fs_exclude_paths": ("CLIPPY_FS_EXCLUDE_PATHS", ["CMakeFiles"]),
}



# Copyright 2020 Lawrence Livermore National Security, LLC and other CLIPPy Project Developers.
# See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
"""
Defines types and classes for Clippy.
"""

import os
from typing import Any

from .error import ClippyConfigurationError

# AnyDict is a convenience type so we can find places
# to be more specific in the future.
AnyDict = dict[str, Any]

# CONFIG_ENTRY is a convenience type for use in CLIPPY_CONFIG.
CONFIG_ENTRY = tuple[str | None, Any]


# CLIPPY_CONFIG holds configuration items for both
# global settings as well as backend settings.
# All access to config should be via the `get()` method.
class CLIPPY_CONFIG:
    def __init__(self, d):
        self._entries = d

    def get(self, field) -> Any:
        if field not in self._entries:
            raise ClippyConfigurationError(f"unknown configuration setting {field}")
        env, val = self._entries[field]
        if env is None:
            return val
        return os.environ.get(env, val)



"""
This module holds backend interfaces to clippy.

All backends must expose a `classes()` method that
returns a dict of {name: class} to import into clippy and
a `get_cfg()` method that returns a CLIPPY_CONFIG instance.
"""
